\ The PPC assembler.
\
\ Copyright (C) 2006, 2007  Joshua I. Grams
\ May be used, modified, and redistributed without limitation.

require modules.f

module assembler

public:

	: assembler [expose-module] assembler ;

private:

module (assembler)

\ Registers

\ magic numbers, randomly chosen
$E027CBA0 constant gpr
$FA1987E0 constant fpr
: is-gpr? ( x -- flag ) gpr dup 32 + within ;
: is-fpr? ( x -- flag ) fpr dup 32 + within ;
: gpr> ( x -- gpr ) gpr - ;
: fpr> ( x -- fpr ) fpr - ;

: enum ( n "name" -- n+1 ) dup constant 1+ ;

public: 

	gpr enum r0  enum r1  enum r2  enum r3
		enum r4  enum r5  enum r6  enum r7
		enum r8  enum r9  enum r10 enum r11
		enum r12 enum r13 enum r14 enum r15
		enum r16 enum r17 enum r18 enum r19
		enum r20 enum r21 enum r22 enum r23
		enum r24 enum r25 enum r26 enum r27
		enum r28 enum r29 enum r30 enum r31 drop

	fpr enum fr0  enum fr1  enum fr2  enum fr3
		enum fr4  enum fr5  enum fr6  enum fr7
		enum fr8  enum fr9  enum fr10 enum fr11
		enum fr12 enum fr13 enum fr14 enum fr15
		enum fr16 enum fr17 enum fr18 enum fr19
		enum fr20 enum fr21 enum fr22 enum fr23
		enum fr24 enum fr25 enum fr26 enum fr27
		enum fr28 enum fr29 enum fr30 enum fr31 drop

private:

\ flags (extensions to the mnemonic)
variable flags  0 flags !
: +flags ( instr -- instr' )  flags @ or  0 flags ! ;

variable stem-xt
create stem-buf 2 c, char : c, bl c, 33 chars allot

: append ( c-addr1 u1 c-addr2 u2 -- c-addr1 u3 )
	dup >r  2over chars + swap chars move  r> + ;

: save-stem ( >in -- )
	dup >in ! ' stem-xt !  >in !
	stem-buf char+ 2 parse-word 1- append stem-buf c! drop ;

: stem ( -- c-addr u )  stem-buf count 2 /string ;

\ takes a suffix and starts a colon definition for stem+suffix.
: +define ( c-addr u -- )  stem-buf count 2swap append evaluate ;

wordlist constant *flags

: set-flags ( >in xt -- )
	>r >in @  over save-stem  swap >in !
	get-current *flags set-current r@ alias  set-current  >in !
	r> execute ;

: find-flags ( c-addr u -- xt )
	*flags search-wordlist -1 <> abort" mnemonic has no flags" ;

: 'flags ( "name" -- xt )  parse-word find-flags ;

: suffix ( c-addr u bits -- )
	>r +define r> postpone literal postpone flags postpone +!
	stem-xt @ compile, postpone ; ;

: no-flags ;

: o-flag s" o," $400 suffix ;
: .-flag s" .," 1 suffix ;
: o.-flags s" o.," $401 suffix  o-flag .-flag ;

: l-flag s" l," 1 suffix ;
: a-flag s" a," 2 suffix ;
: la-flags s" la," 3 suffix  l-flag a-flag ;


\ the basic divisions: all are ( u -- u' )
: F0 $3f and 26 lshift ;
: F1 $1f and 21 lshift ;
: F2 $1f and 16 lshift ;
: F3 $1f and 11 lshift ;
: F4 $1f and 6 lshift ;  : F4' $f and 6 lshift ;
: F5 $1f and 1 lshift ;

\ bitfields.  ( u code -- code' ) unless otherwise marked.
: OP swap F0 or ;
: XO swap F5 or ;
: XO' swap F4 or ;  : aXO' swap F4' or ;
: OPs ( OP XO XO' -- ) XO' XO OP ;
: aOPs ( aXO' XO OP -- ) aXO' XO OP ;

: rT swap dup is-gpr? 0= abort" invalid rT"  gpr> F1 or ;
: rA swap dup is-gpr? 0= abort" invalid rA"  gpr> F2 or ;
: rA|0 swap  dup 0= if 0 else gpr> dup 0= then abort" invalid rA|0"  F2 or ;
: rA+ swap  gpr> dup 0= abort" invalid rA+"  F2 or ;
: rB swap dup is-gpr? 0= abort" invalid rB"  gpr> F3 or ;
: rC swap dup is-gpr? 0= abort" invalid rC"  gpr> F4 or ;

: frT swap dup is-fpr? 0= abort" invalid frT" fpr> F1 or ;
: frA swap dup is-fpr? 0= abort" invalid frA" fpr> F2 or ;
: frB swap dup is-fpr? 0= abort" invalid frB" fpr> F3 or ;
: frC swap dup is-fpr? 0= abort" invalid frC" fpr> F4 or ;

: lu-check ( rT rA -- rT rA )  2dup = abort" can't load and update same register" ;

: crbD swap F1 or ;
: crbA swap F2 or ;
: crbB swap F3 or ;

: SIMM ( n code -- code' )
	swap  dup $8000 and 1 lshift +  dup 0 $10000 within
	0= abort" invalid signed 16-bit number" or ;
: UIMM ( u code -- code' )
	over 0 $10000 within 0= abort" invalid unsigned 16-bit number" or ;
: A16 ( addr code -- code' )
	tuck 2 and 0= if here - then dup $8000 and 1 lshift +
	dup 3 and abort" branch target not aligned"
	dup $ffff0000 and abort" branch target out of range" or ;
: A26 ( addr code -- code' )
	tuck 2 and 0= if here - then dup $2000000 and 1 lshift +
	dup 3 and abort" branch target not aligned"
	dup $fc000000 and abort" branch target out of range" or ;
: BC swap dup $1f u> abort" invalid branch condition" F1 or ;
: BC~CTR swap dup $1f u> over 6 and 4 <> or abort" invalid bcctr condition" F1 or ;
: TC swap dup $1f u> abort" invalid trap condition" F1 or ;
: BB swap F2 or ;
: CRM swap $ff and 12 lshift or ;
: spr swap dup $1f and 16 lshift swap $3e0 and 6 lshift or or ;
' spr alias tbr
: SR swap $f and 16 lshift or ;
: FM swap $ff and 17 lshift or ;
: L swap 1 and 21 lshift or ;
: IMM swap 1 lshift F3 or ;
: N swap F3 or ;
: MB swap F4 or ;
: ME swap F5 or ;


variable #crfs 0 #crfs !
: CRF ( u -- u' ) 2 lshift  1 #crfs +! ;
: -CRF -1 #crfs +! ;
: ?CRF #crfs @ if -CRF else 0 then ;
: crfD -CRF swap F1 or ;  : ?crfD >r ?CRF r> swap F1 or ;
: crfS -CRF swap F2 or ;

public:

	: CR0 0 CRF ;  : CR1 1 CRF ;  : CR2 2 CRF ;  : CR3 3 CRF ;
	: CR4 4 CRF ;  : CR5 5 CRF ;  : CR6 6 CRF ;  : CR7 7 CRF ;

	: LT ( crf -- crb ) ?CRF 0 or ;
	: GT ( crf -- crb ) ?CRF 1 or ;
	: EQ ( crf -- crb ) ?CRF 2 or ;
	: SO ( crf -- crb ) ?CRF 3 or ;
	4 constant FX
	5 constant FEX
	6 constant VX
	7 constant OX

	\ trap condition bits
	31 constant T*
	16 constant T<
	8  constant T>
	4  constant T=
	2  constant TU<
	1  constant TU>

private:


\ Defining Instruction Forms

create buf 33 chars allot

: flags-xt ( c-addr u -- xt )
	buf place  buf count 2 min
	dup 1 > if s" -flags" else s" -flag" then append buf c! drop
	buf find -1 <> abort" unknown flags" ;

: :op ( opcodes "name" -- >in colon-sys )
	>in @ swap >r : r> postpone literal postpone +flags ;

: ;op ( colon-sys -- )  postpone , postpone ; ; immediate

: postpone-;op postpone ;op ;


: :form ( "name" -- )  : 0 postpone literal ;

\ parse the next word of input, then reset >IN.
: peek-word ( -- c-addr u )  >in @  parse-word  rot >in ! ;

: takes> ( "flags ... ;" -- )
	parse-word flags-xt  postpone :op
	begin peek-word s" ;" compare while postpone postpone repeat
	postpone postpone-;op  postpone literal postpone set-flags ; immediate


\ Instruction forms

:form no-args OPs takes> none ; \ no arguments

:form arith-i \ arithmetic with immediate
	OP takes> none SIMM rA rT ;
:form arith-2 \ arithmetic with two arguments
	aOPs takes> o. rA rT ;
:form arith-3 \ arithmetic with three arguments
	aOPs takes> o. rB rA rT ;
:form mulhw aOPs takes> . rB rA rT ;

:form logical-i OP takes> none UIMM rT rA ;
:form logical-2 OPs takes> . rT rA ;
:form logical-3 OPs takes> . rB rT rA ;
:form rotate OP takes> . ME MB rB rT rA ;
:form rotate-i OP takes> . ME MB N rT rA ;
:form srawi OPs takes> . N rT rA ;

:form float-2 OPs takes> . frB frT ;
:form float-3 OPs takes> . frB frA frT ;
:form float-3* XO OP takes> . frC frA frT ;
:form float-4 XO OP takes> . frB frC frA frT ;

:form integer OP takes> none SIMM rA|0 rT ;
:form indexed OPs takes> none rB rA|0 rT ;
:form updating-store OP takes> none SIMM rA+ rT ;
:form updating-load OP takes> none SIMM lu-check rA+ rT ;
:form indexed-updating-store OPs takes> none rB rA+ rT ;
:form indexed-updating-load OPs takes> none rB lu-check rA+ rT ;

:form float OP takes> none SIMM rA|0 frT ;
:form updating-float OP takes> none SIMM rA+ frT ;
:form indexed-float OPs takes> none rB rA|0 frT ;
:form indexed-updating-float OPs takes> none rB rA+ frT ;

:form stwcx. OPs 1 or takes> none rB rA|0 rT ;
:form cache OPs takes> none rB rA|0 ;
:form string OPs takes> none N rA|0 rT ;

:form compare OPs takes> none rB rA L ?crfD ;
:form cmpi OP takes> none SIMM rA L ?crfD ;
:form cmpli OP takes> none UIMM rA L ?crfD ;
:form float-compare OPs takes> none frB frA crfD ;

:form cond-bit OPs takes> none crbB crbA crbD ;
:form mcrxr OPs takes> none crfD ;
:form cond-field OPs takes> none crfS crfD ;
:form mtcrf OPs takes> none rT CRM ;

:form mtfsb OPs takes> . crbD ;
:form mtfsf OPs takes> . frB FM ;
:form mtfsfi OPs takes> . IMM crfD ;
:form mffs OPs takes> . frT ;

:form branch OP takes> la A26 ;
:form ?branch OP takes> la A16 BB BC ;
:form bclr OPs takes> l  BB BC ;
:form bcctr OPs takes> l BB BC~CTR ;
:form trap OPs takes> none rB rA TC ;
:form trap-i OP takes> none SIMM rA TC ;

\ segment registers
:form mfsr OPs takes> none SR rT ;
:form mtsr OPs takes> none rT SR ;
:form srin OPs takes> none rB rT ;

:form special-register OPs takes> none rT ;
:form mtspr OPs takes> none rT spr ;
:form mfspr OPs takes> none spr rT ;
:form mftbr OPs takes> none tbr rT ;

:form tlbie OPs takes> none rB ;

public:

	include instructions.f

private:


\ Defining simplified mnemonics

: :asm ( "name" -- >in colon-sys )  >in @ : ;

: ;extends ( >in colon-sys "name" -- )
	>in @ >r  ' compile, postpone ;
	r> >in ! 'flags set-flags ; immediate

public:

	include simplified.f

	: ext16 ( n -- n' )  16 lshift 65536 / ;

	: split ( n -- n@ha n@l )
		dup $8000 and 1 lshift +  dup 16 rshift  swap ext16 ;

end-module \ (assembler)

end-module \ assembler
